
from functools import lru_cache

"""
if installed, eth_portfolio extends the db with some extra stuffs and we want to insert the objects to the db using the extended classes.
The functions in this file lets us do that.
"""

@lru_cache(maxsize=1)
def _get_get_block():
    try:
        from eth_portfolio._db.utils import get_block
    except ImportError:
        from y._db.utils import get_block
    return get_block
    
@lru_cache(maxsize=1)
def _get_get_token():
    try:
        from eth_portfolio._db.utils import get_token
    except ImportError:
        from y._db.utils.token import get_token
    return get_token



import logging
import os
from typing import Any, Callable

import eth_retry
from brownie import chain, web3
from requests import Session
from requests.adapters import HTTPAdapter
from web3 import HTTPProvider, Web3
from web3.middleware import filter
from web3.middleware.geth_poa import geth_poa_middleware

from y import ENVIRONMENT_VARIABLES as ENVS
from y.networks import Network
from y.utils.cache import memory

logger = logging.getLogger(__name__)

provider_specific_batch_sizes = {
    "moralis":  2_000,
    "pokt":     2_000,
    "tenderly": 2_000,
    "ankr":     2_000,
}

chain_specific_max_batch_sizes = {
    Network.Mainnet:    10_000,     # 1.58 days
    Network.Gnosis:     20_000,     # 1.15 days
    Network.Fantom:     100_000,    # 1.03 days
    Network.Arbitrum:   20_000,     # 0.34 days
    Network.Optimism:   800_000,    # 10.02 days
}

fallback_batch_size = 10_000

def _get_batch_size() -> int:
    if batch_size := ENVS.GETLOGS_BATCH_SIZE:
        return batch_size
    for provider, size in provider_specific_batch_sizes.items():
        if provider in web3.provider.endpoint_uri:
            return size
    return chain_specific_max_batch_sizes.get(chain.id, fallback_batch_size)

BATCH_SIZE = _get_batch_size()


def should_cache(method: str, params: Any) -> bool:
    if method == "eth_getCode" and params[1] == "latest":
        return True
    return False


def cache_middleware(make_request: Callable, web3: Web3) -> Callable:
    @eth_retry.auto_retry
    def middleware(method: str, params: Any) -> Any:
        logger.debug("%s %s", method, params)
        if should_cache(method, params):
            return memory.cache(make_request)(method, params)
        return make_request(method, params)
    return middleware


def setup_middleware() -> None:
    # patch web3 provider with more connections and higher timeout
    if web3.provider:
        try:
            assert web3.provider.endpoint_uri.startswith("http"), "only http and https providers are supported"
            adapter = HTTPAdapter(pool_connections=100, pool_maxsize=100)
            session = Session()
            session.mount("http://", adapter)
            session.mount("https://", adapter)
            web3.provider = HTTPProvider(web3.provider.endpoint_uri, {"timeout": 600}, session)
        except AttributeError as e:
            if "'IPCProvider' object has no attribute 'endpoint_uri'" in str(e):
                pass 
            else:
                raise

    # patch and inject local filter middleware
    filter.MAX_BLOCK_REQUEST = BATCH_SIZE
    web3.middleware_onion.add(filter.local_filter_middleware)
    web3.middleware_onion.add(cache_middleware)
    
    if chain.id == Network.Optimism:
        web3.middleware_onion.inject(geth_poa_middleware, layer=0)

def ensure_middleware() -> None:
    setup_middleware()
    




from pony.orm import BindingError, TransactionError

from y._db.config import connection_settings
from y._db.entities import db
from y._db.utils.utils import get_chain, get_block

        
try:
    db.bind(**connection_settings)
    db.generate_mapping(create_tables=True)
except TransactionError as e:
    if str(e) != "@db_session-decorated create_tables() function with `ddl` option cannot be called inside of another db_session":
        raise e
except BindingError as e:
    if not str(e).startswith('Database object was already bound to'):
        raise e

__all__ = 'get_chain', 'get_block'



from y import ENVIRONMENT_VARIABLES as ENVS
import errno
from os import mkdir, path

if ENVS.DB_PROVIDER == "sqlite":
    try:
        mkdir(f"{path.expanduser( '~' )}/.ypricemagic")
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    connection_settings = {
        'provider': str(ENVS.DB_PROVIDER),
        'filename': f"{path.expanduser( '~' )}/.ypricemagic/ypricemagic.sqlite",
        'create_db': True,
    }
else:
    connection_settings = {
        'provider': str(ENVS.DB_PROVIDER),
        'host': str(ENVS.DB_HOST),
        'user': str(ENVS.DB_USER),
        'password': str(ENVS.DB_PASSWORD),
        'database': str(ENVS.DB_DATABASE),
    }
    if ENVS.DB_PORT:
        connection_settings['port'] = int(ENVS.DB_PORT)




import logging
import time
from functools import lru_cache, wraps
from typing import Callable, TypeVar
from typing_extensions import ParamSpec

from a_sync import a_sync
from a_sync.primitives.executor import PruningThreadPoolExecutor
from pony.orm import (CommitException, OperationalError, TransactionError,
                      UnexpectedError, commit, db_session)

_T = TypeVar('_T')
_P = ParamSpec('_P')

logger = logging.getLogger(__name__)


token_attr_threads = PruningThreadPoolExecutor(32)

def retry_locked(callable: Callable[_P, _T]) -> Callable[_P, _T]:
    @wraps(callable)
    def retry_locked_wrap(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        sleep = 0.05
        while True:
            try:
                retval = callable(*args, **kwargs)
                commit()
                return retval
            except (CommitException, OperationalError, UnexpectedError) as e:
                logger.debug("%s.%s got exc %s", callable.__module__, callable.__name__, e)
                if "database is locked" not in str(e):
                    raise e
                time.sleep(sleep)
                sleep *= 1.5
            except TransactionError as e:
                logger.debug("%s.%s got exc %s", callable.__module__, callable.__name__, e)
                if "An attempt to mix objects belonging to different transactions" not in str(e):
                    raise e
    return retry_locked_wrap

a_sync_db_session = lambda fn: a_sync(default='async', executor=token_attr_threads)(
    db_session(
        retry_locked(
            fn
        )
    )
)
a_sync_db_session_cached = lambda fn: a_sync(default='async', executor=token_attr_threads)(
    retry_locked(
        lru_cache(maxsize=None)(
            db_session(
                fn
            )
        )
    )
)


# async: done

import asyncio
import logging
from typing import Optional

from brownie import chain

from y.classes.common import ERC20
from y.constants import wbtc, weth
from y.contracts import Contract
from y.exceptions import NonStandardERC20
from y.networks import Network
from y.prices.lending.aave import aave
from y.prices.lending.compound import CToken
from y.prices.stable_swap.curve import CurvePool
from y.prices.utils.buckets import check_bucket
from y.prices.yearn import YearnInspiredVault

logger = logging.getLogger(__name__)

# This module is far from perfect, but provides an acceptable way to validate some of the prices returned by `get_price`

acceptable_all_chains = [
    weth.address,
    wbtc.address,
]
ACCEPTABLE_HIGH_PRICES = {
    Network.Mainnet: [
        # eth and eth-like
        "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE", # eth
        "0x7f39C581F595B53c5cb19bD0b3f8dA6c935E2Ca0", # wsteth
        "0xA3D87FffcE63B53E0d54fAa1cc983B7eB0b74A9c", # ecrv
        "0xaA17A236F2bAdc98DDc0Cf999AbB47D47Fc0A6Cf", # ankrcrv
        "0x53a901d48795C58f485cBB38df08FA96a24669D5", # rcrv
        "0xC4C319E2D4d66CcA4464C0c2B32c9Bd23ebe784e", # alETH+ETH-f
        "0x06325440D014e39736583c165C2963BA99fAf14E", # stecrv
        "0x5e74C9036fb86BD7eCdcb084a0673EFc32eA31cb", # sETH
        "0xae78736Cd615f374D3085123A210448E74Fc6393", # rETH
        "0xE95A203B1a91a908F9B9CE46459d101078c2c3cb", # aETHc
        "0xae7ab96520DE3A18E5e111B5EaAb095312D7fE84", # stETH
        "0x9559Aaa82d9649C7A7b220E7c461d2E74c9a3593", # rETH
        "0x836A808d4828586A69364065A1e064609F5078c7", # pETH
        "0xc3D088842DcF02C13699F936BB83DFBBc6f721Ab", # vETH
        "0xC1330aCBbcE97cb9695B7ee161c0F95B875a8b0F", # onETH
        "0x5E8422345238F34275888049021821E8E08CAa1f", # fxsETH
        "0x856c4Efb76C1D1AE02e20CEB03A2A6a08b0b8dC3", # OETH
        "0xE72B141DF173b999AE7c1aDcbF60Cc9833Ce56a8", # ETH+
        "0x3d1E5Cf16077F349e999d6b21A4f646e83Cd90c5", # dETH
        "0x0100546F2cD4C9D97f798fFC9755E47865FF7Ee6", # alETH
        "0xBe9895146f7AF43049ca1c1AE358B0541Ea49704", # cbETH
        "0x7C07F7aBe10CE8e33DC6C5aD68FE033085256A84", # icETH
        "0xA35b1B31Ce002FBF2058D22F30f95D405200A15b", # ETHx
        "0x64351fC9810aDAd17A690E4e1717Df5e7e085160", # msETH
        "0x821A278dFff762c76410264303F25bF42e195C0C", # pETH
        "0xa2E3356610840701BDf5611a53974510Ae27E2e1", # wbETH
        "0x1BED97CBC3c24A4fb5C069C6E311a967386131f7", # yETH
        # btc and btc-like
        "0xEB4C2781e4ebA804CE9a9803C67d0893436bB27D", # renbtc
        "0xfE18be6b3Bd88A2D2A7f928d00292E7a9963CfC6", # sbtc
        "0x49849C98ae39Fff122806C06791Fa73784FB3675", # crvrenwbtc
        "0x075b1bb99792c9E1041bA13afEf80C91a1e70fB3", # crvrenwsbtc
        "0xb19059ebb43466C323583928285a49f558E572Fd", # hcrv
        "0x64eda51d3Ad40D56b9dFc5554E06F94e1Dd786Fd", # tbtc/sbtcCrv
        "0xDE5331AC4B3630f94853Ff322B66407e0D6331E8", # pBTC/sbtcCRV
        "0x410e3E86ef427e30B9235497143881f717d93c2A", # bBTC/sbtcCRV
        "0x2fE94ea3d5d4a175184081439753DE15AeF9d614", # oBTC/sbtcCRV
        "0x0327112423F3A68efdF1fcF402F6c5CB9f7C33fd", # btc++
        "0x3212b29E33587A00FB1C83346f5dBFA69A458923", # imbtc
        "0x5228a22e72ccC52d415EcFd199F99D0665E7733b", # pbtc
        "0xFbdCA68601f835b27790D98bbb8eC7f05FDEaA9B", # ibbtc/sbtcCRV-f
        "0x0316EB71485b0Ab14103307bf65a021042c6d380", # hbtc
        "0x9BE89D2a4cd102D8Fecc6BF9dA793be995C22541", # bbtc
        "0x8064d9Ae6cDf087b1bcd5BDf3531bD5d8C537a68", # obtc
        "0x8dAEBADE922dF735c38C80C7eBD708Af50815fAa", # tbtc
        "0x8751D4196027d4e6DA63716fA7786B5174F04C15", # wibbtc
        "0x66eFF5221ca926636224650Fd3B9c497FF828F7D", # multiBTC
        "0x18084fbA666a33d37592fA2633fD49a74DD93a88", # tbtc
        # gold tokens
        "0x45804880De22913dAFE09f4980848ECE6EcbAf78", # paxg
        "0x4922a015c4407F87432B179bb209e125432E4a2A", # xaut
        # other
        "0x0bc529c00C6401aEF6D220BE8C6Ea1667F6Ad93e", # yfi
        "0x1cEB5cB57C4D4E2b2433641b95Dd330A33185A44", # kp3r
        "0xD5525D397898e5502075Ea5E830d8914f6F0affe", # meme
        "0x9f8F72aA9304c8B593d555F12eF6589cC3A579A2", # mkr
        "0x3A283D9c08E8b55966afb64C515f5143cf907611", # crvCVXETH
        "0x23B608675a2B2fB1890d3ABBd85c5775c51691d5", # socks
        "0xcA3d75aC011BF5aD07a98d02f18225F9bD9A6BDF", # crvtricrypto
        "0xc4AD29ba4B3c580e6D59105FFf484999997675Ff", # crv3crypto
        "0xd075e95423C5c4BA1E122CaE0f4CdFA19b82881b", # wPE
        "0xe9F84dE264E91529aF07Fa2C746e934397810334", # sak3
        "0xa1d0E215a23d7030842FC67cE582a6aFa3CCaB83", # yfii
        "0x0fe629d1E84E171f8fF0C1Ded2Cc2221Caa48a3f", # mask
        "0x5F0E628B693018f639D10e4A4F59BD4d8B2B6B44", # white
        "0xdBdb4d16EdA451D0503b854CF79D55697F90c8DF", # alcx
        "0x3aaDA3e213aBf8529606924d8D1c55CbDc70Bf74", # xMON
        "0x892A6f9dF0147e5f079b0993F486F9acA3c87881", # xFUND
        "0x0ab87046fBb341D058F17CBC4c1133F25a20a52f", # gOHM
        "0x97983236bE88107Cc8998733Ef73D8d969c52E37", # sdYFI
        "0x68749665FF8D2d112Fa859AA293F07A622782F38", # XAUt
        "0xf5f5B97624542D72A9E06f04804Bf81baA15e2B4", # crvUSDTWBTCWETH
        "0x7F86Bf177Dd4F3494b841a37e810A34dD56c829B", # crvUSDCWBTCWETH
        "0x2889302a794dA87fBF1D6Db415C1492194663D13", # crvCRVUSDTBTCWSTETH
        "0xBfAb6FA95E0091ed66058ad493189D2cB29385E6", # ETHwBETHCRV
        # nfts
        "0x641927E970222B10b2E8CDBC96b1B4F427316f16", # meeb
        "0x9cea2eD9e47059260C97d697f82b8A14EfA61EA5", # punk
        "0x269616D549D7e8Eaa82DFb17028d0B212D11232A", # punk (different)
        "0x69BbE2FA02b4D90A944fF328663667DC32786385", # punk-basic
        "0xD70240Dd62F4ea9a6A2416e0073D72139489d2AA", # glyph
        "0x114f1388fAB456c4bA31B1850b244Eedcd024136", # coolcats
        "0xEA47B64e1BFCCb773A0420247C0aa0a3C1D2E5C5", # bayc
    ],
    Network.BinanceSmartChain: [
        # eth and eth-like
        "0x8d0e18c97e5dd8ee2b539ae8cd3a3654df5d79e5", # bweth
        "0xbfF4a34A4644a113E8200D7F1D79b3555f723AfE", # ibeth
        # btc and btc-like
        "0x7130d2A12B9BCbFAe4f2634d864A1Ee1Ce3Ead9c", # btcb
        "0x08FC9Ba2cAc74742177e0afC3dC8Aed6961c24e7", # ibbtcb
    ],
    Network.Fantom: [
        # eth and eth-like
        "0xBDC8fd437C489Ca3c6DA3B5a336D11532a532303", # anyeth
        # btc and btc-like
        "0xDBf31dF14B66535aF65AaC99C32e9eA844e14501", # renbtc
        "0x2406dCe4dA5aB125A18295f4fB9FD36a0f7879A2", # anybtc
        # other
        "0x29b0Da86e484E1C0029B56e817912d778aC0EC69", # yfi
        "0xf43Cc235E686d7BC513F53Fbffb61F760c3a1882", # elite
        "0x58e57cA18B7A47112b877E31929798Cd3D703b0f", # crv3crypto
    ],
    Network.Avalanche: [
        # btc and btc-like
        "0xDBf31dF14B66535aF65AaC99C32e9eA844e14501", # renbtc
        # other
        "0xd6070ae98b8069de6B494332d1A1a81B6179D960", # bifi
    ],
    Network.Optimism: [
        # eth and eth-like
        "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE", # eth
        "0x9Bcef72be871e61ED4fBbc7630889beE758eb81D", # reth
        "0x1F32b1c2345538c0c6f582fCB022739c4A194Ebb", # wsteth
        "0x6806411765Af15Bddd26f8f544A34cC40cb9838B", # frxeth
        "0x484c2D6e3cDd945a8B2DF735e079178C1036578c", # sfrxeth
        "0x1610e3c85dd44Af31eD7f33a63642012Dca0C5A5", # mseth
        "0x3E29D3A9316dAB217754d13b28646B76607c5f04", # aleth
        "0xE405de8F52ba7559f9df3C368500B6E6ae6Cee49", # seth
        # btc and btc-like
        "0x298B9B95708152ff6968aafd889c6586e9169f1D", # sbtc
        "0x6c84a8f1c29108F47a79964b5Fe888D4f4D0dE40", # tbtc
    ],
    Network.Arbitrum: [
        "0x8e0B8c8BB9db49a46697F3a5Bb8A308e744821D2", # crv3crypto
    ],
    Network.Base: [
        # eth and eth-like
        "0x2Ae3F1Ec7F1F5012CFEab0185bfc7aa3cf0DEc22", # cbeth
        # btc and btc-like
        "0x236aa50979D5f3De3Bd1Eeb40E81137F22ab794b", # tbtc
    ],
}.get(chain.id, []) + acceptable_all_chains

async def _sense_check(
    token_address: str, 
    block: Optional[int],
    price: float
    ):

    # if price is in "normal" range, exit sense check
    if price < 1000:
        return None
    
    # if we've already validated that the token should have a high price, exit sense check
    if token_address in ACCEPTABLE_HIGH_PRICES:
        return None

    # for some token types, its normal to have a crazy high nominal price
    # we can skip the sense check for those
    if await _exit_sense_check(token_address):
        return None

    # proceed with sense check
    price_readable = round(price, 4)
    network = Network.name(chain.id)
    try:
        symbol = await ERC20(token_address, asynchronous=True).symbol
        logger.warning(f'unusually high price (${price_readable}) returned for {symbol} {token_address} on {network} block {block}. This does not necessarily mean that the price is wrong, but you may want to validate the price for yourself before proceeding.')
    except NonStandardERC20:
        logger.warning(f'unusually high price (${price_readable}) returned for {token_address} on {network} block {block}. This does not necessarily mean that the price is wrong, but you may want to validate the price for yourself before proceeding.')


#yLazyLogger(logger)
async def _exit_sense_check(token_address: str) -> bool:
    '''
    For some token types, its normal to have a crazy high nominal price.
    We can skip the sense check for those.
    We can also skip wrapped versions of tokens in `ACCEPTABLE_HIGH_PRICES`.
    '''

    bucket = await check_bucket(token_address, sync=False)

    if bucket in ['uni or uni-like lp', 'balancer pool']:
        return True

    elif bucket == 'curve lp':
        underlyings = await CurvePool(token_address, asynchronous=True).coins
        if questionable_underlyings := [und for und in underlyings if und not in ACCEPTABLE_HIGH_PRICES]:
            return all(await asyncio.gather(*[_exit_sense_check(underlying) for underlying in questionable_underlyings]))
        return True
    
    elif bucket == 'atoken':
        underlying = await aave.underlying(token_address, sync=False)
    elif bucket == 'compound':
        underlying = await CToken(token_address, asynchronous=True).underlying
    elif bucket == 'solidex':
        contract = await Contract.coroutine(token_address)
        underlying = await contract.pool.coroutine()
    elif bucket == 'yearn or yearn-like':
        underlying = await YearnInspiredVault(token_address, asynchronous=True).underlying
    else:
        return False

    return underlying in ACCEPTABLE_HIGH_PRICES or await _exit_sense_check(underlying)



import asyncio
import logging
from decimal import Decimal
from typing import Optional

import a_sync
from brownie import chain

from y import convert, Contract
from y.constants import weth
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.prices import magic
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

class wstEth(a_sync.ASyncGenericBase):
    def __init__(self, asynchronous: bool = False) -> None:
        self.asynchronous = asynchronous
        try:
            self.address = {
                Network.Mainnet: '0x7f39C581F595B53c5cb19bD0b3f8dA6c935E2Ca0'
            }[chain.id]
            self.wrapped_for_curve = {
                Network.Mainnet: '0xb82CFa4325568748506dC7cF267857Ff1e3b8d39'
            }[chain.id]
        except KeyError:
            self.address = None

    async def get_price(self, block: Optional[Block] = None) -> UsdPrice:
        share_price, weth_price = await asyncio.gather(
            raw_call(self.address, 'stEthPerToken()', output='int', block=block, sync=False),
            magic.get_price(weth, block, sync=False),
        )
        share_price /= Decimal(10 ** 18)
        return UsdPrice(share_price * Decimal(float(weth_price)))

wsteth = wstEth(asynchronous=True)

def is_wsteth(address: AnyAddressType) -> bool:
    return chain.id == Network.Mainnet and convert.to_address(address) in [wsteth.address, wsteth.wrapped_for_curve]




from functools import lru_cache
from typing import List, Optional

from dank_mids.types import _DictStruct
from inflection import underscore


class _CamelDictStruct(_DictStruct, rename="camel"):
    """
    A mixin class that allows camelCase key lookups for snake_case struct attrs
    Original use case was so Log structs can be used interchangably with LogReceipt instances
    """
    def __getitem__(self, attr: str):
        return getattr(self, _make_snake(attr))

class Log(_CamelDictStruct):
    removed: Optional[bool]
    log_index: Optional[int]
    transaction_index: Optional[int]
    transaction_hash: str
    block_hash: Optional[str]
    block_number: Optional[int]
    address: Optional[str]
    data: Optional[str]
    topics: Optional[List[str]]

class Trace(_CamelDictStruct):
    # TODO so we can trace chain for eth_port on alt chains
    pass

@lru_cache(maxsize=None)
def _make_snake(camel: str) -> str:
    return underscore(camel)



import logging
from datetime import datetime, timezone
from functools import lru_cache
from typing import Optional

from brownie import chain

from y._db.entities import Block, Chain, insert
from y._db.utils._ep import _get_get_block
from y._db.utils.decorators import a_sync_db_session, a_sync_db_session_cached

logger = logging.getLogger(__name__)


@a_sync_db_session
def get_chain() -> Chain:
    return Chain.get(id=chain.id) or insert(type=Chain, id=chain.id) or Chain[chain.id]

@lru_cache
def ensure_chain() -> None:
    """ensures that the chain object for the connected chain has been inserted to the db"""
    get_chain(sync=True)

@a_sync_db_session
def get_block(number: int) -> Block:
    ensure_chain()
    if block := Block.get(chain=chain.id, number=number):
        return block
    return insert(type=Block, chain=chain.id, number=number) or get_block(number, sync=True)

@a_sync_db_session_cached
def ensure_block(number: int) -> None:
    get_block = _get_get_block()
    get_block(number, sync=True)

@a_sync_db_session
def get_block_timestamp(number: int) -> Optional[int]:
    get_block = _get_get_block()
    block = get_block(number, sync=True)
    if ts := block.timestamp:
        unix = ts.timestamp()
        logger.debug("got %s.timestamp from cache: %s, %s", block, unix, ts)
        return unix
    
@a_sync_db_session
def set_block_timestamp(block: int, timestamp: int) -> None:
    get_block = _get_get_block()
    block = get_block(block, sync=True)
    timestamp = datetime.fromtimestamp(timestamp, tz=timezone.utc)
    block.timestamp = timestamp
    logger.debug("cached %s.timestamp %s", block, timestamp)



UNIV3_QUOTER_ABI = [{"inputs":[{"internalType":"address","name":"_factory","type":"address"},{"internalType":"address","name":"_WETH9","type":"address"}],"stateMutability":"nonpayable","type":"constructor"},{"inputs":[],"name":"WETH9","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[],"name":"factory","outputs":[{"internalType":"address","name":"","type":"address"}],"stateMutability":"view","type":"function"},{"inputs":[{"internalType":"bytes","name":"path","type":"bytes"},{"internalType":"uint256","name":"amountIn","type":"uint256"}],"name":"quoteExactInput","outputs":[{"internalType":"uint256","name":"amountOut","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"tokenIn","type":"address"},{"internalType":"address","name":"tokenOut","type":"address"},{"internalType":"uint24","name":"fee","type":"uint24"},{"internalType":"uint256","name":"amountIn","type":"uint256"},{"internalType":"uint160","name":"sqrtPriceLimitX96","type":"uint160"}],"name":"quoteExactInputSingle","outputs":[{"internalType":"uint256","name":"amountOut","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"bytes","name":"path","type":"bytes"},{"internalType":"uint256","name":"amountOut","type":"uint256"}],"name":"quoteExactOutput","outputs":[{"internalType":"uint256","name":"amountIn","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"address","name":"tokenIn","type":"address"},{"internalType":"address","name":"tokenOut","type":"address"},{"internalType":"uint24","name":"fee","type":"uint24"},{"internalType":"uint256","name":"amountOut","type":"uint256"},{"internalType":"uint160","name":"sqrtPriceLimitX96","type":"uint160"}],"name":"quoteExactOutputSingle","outputs":[{"internalType":"uint256","name":"amountIn","type":"uint256"}],"stateMutability":"nonpayable","type":"function"},{"inputs":[{"internalType":"int256","name":"amount0Delta","type":"int256"},{"internalType":"int256","name":"amount1Delta","type":"int256"},{"internalType":"bytes","name":"path","type":"bytes"}],"name":"uniswapV3SwapCallback","outputs":[],"stateMutability":"view","type":"function"}]


import asyncio
import logging
from typing import Optional

import a_sync
from brownie import chain

from y import convert
from y.classes.common import ERC20
from y.constants import weth
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.prices import magic
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

def is_creth(token: AnyAddressType) -> bool:
    address = convert.to_address(token)
    return chain.id == Network.Mainnet and address == '0xcBc1065255cBc3aB41a6868c22d1f1C573AB89fd'

@a_sync.a_sync(default='sync')
async def get_price_creth(token: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
    address = convert.to_address(token)
    total_balance, total_supply, weth_price = await asyncio.gather(
        raw_call(address, 'accumulated()', output='int', block=block, sync=False),
        ERC20(address, asynchronous=True).total_supply(block),
        magic.get_price(weth, block, sync=False),
    )
    per_share = total_balance / total_supply
    return per_share * weth_price




import logging
import os

from brownie import chain

import y
from y.prices.stable_swap.curve import CurveRegistry, curve

y_logger = logging.getLogger('y')
y_logger.setLevel(logging.DEBUG)
y_logger.addHandler(logging.StreamHandler())

curve: CurveRegistry

def main():
    BAD = os.environ.get("BAD")
    BLOCK = os.environ.get("BLOCK")
    if not BAD:
        raise ValueError("You must specify a token to debug by setting BAD env var")
    if not BLOCK:
        BLOCK = chain.height
        y_logger.warning("no BLOCK specified, using %s", BLOCK)
    if curve.get_pool(BAD, sync=True):
        curve.get_price(BAD, int(BLOCK), sync=True)
    else:
        y_logger.info("%s is not a curve pool", BAD)
    



import asyncio
from typing import Optional, Tuple

import a_sync
from brownie.exceptions import ContractNotFound
from multicall import Call

from y import Contract
from y.classes.common import ERC20, WeiBalance
from y.datatypes import AnyAddressType, Block, UsdPrice, UsdValue
from y.decorators import stuck_coro_debugger
from y.exceptions import ContractNotVerified, MessedUpBrownieContract
from y.utils.cache import memory


@memory.cache()
def is_generic_amm(lp_token_address: AnyAddressType) -> bool:
    try:
        token_contract = Contract(lp_token_address)
        return all(hasattr(token_contract, attr) for attr in ['getReserves','token0','token1'])
    except (ContractNotFound, ContractNotVerified):
        return False
    except MessedUpBrownieContract:
        # probably false, can get more specific when there's a need.
        return False
        
class GenericAmm(a_sync.ASyncGenericBase):
    def __init__(self, asynchronous: bool = False) -> None:
        self.asynchronous = asynchronous

    def __contains__(self, lp_token_address: AnyAddressType) -> bool:
        return is_generic_amm(lp_token_address)
    
    @stuck_coro_debugger
    async def get_price(self, lp_token_address: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
        tvl, total_supply = await asyncio.gather(
            self.get_tvl(lp_token_address, block=block, sync=False),
            ERC20(lp_token_address, asynchronous=True).total_supply_readable(block=block),
        )
        if total_supply is None:
            return None
        elif total_supply == 0:
            return 0
        return UsdPrice(tvl / total_supply)
    
    @stuck_coro_debugger
    @a_sync.a_sync(cache_type='memory')
    async def get_tokens(self, lp_token_address: AnyAddressType) -> Tuple[ERC20,ERC20]:
        tokens = await asyncio.gather(
            Call(lp_token_address, ['token0()(address)']).coroutine(),
            Call(lp_token_address, ['token1()(address)']).coroutine(),
        )
        return ERC20(tokens[0]), ERC20(tokens[1])
    
    @stuck_coro_debugger
    async def get_tvl(self, lp_token_address: AnyAddressType, block: Optional[Block] = None) -> UsdValue:
        lp_token_contract = await Contract.coroutine(lp_token_address)
        tokens = await self.get_tokens(lp_token_address, sync=False)
        reserves = await lp_token_contract.getReserves.coroutine(block_identifier=block)
        reserves = [WeiBalance(reserve,token,block) for token, reserve in zip(tokens,reserves)]
        return UsdValue(sum(await asyncio.gather(*[reserve.__value_usd__(sync=False) for reserve in reserves])))


generic_amm = GenericAmm(asynchronous=True)




import logging
from typing import Optional

from brownie import chain
from y._db.utils._ep import _get_get_token
from y._db.utils.decorators import a_sync_db_session
from y._db.utils.utils import ensure_block


logger = logging.getLogger(__name__)

@a_sync_db_session
def get_deploy_block(address: str) -> Optional[int]:
    get_token = _get_get_token()
    if deploy_block := get_token(address, sync=True).deploy_block:
        logger.debug('%s deploy block from cache: %s', address, deploy_block.number)
        return deploy_block.number
    logger.debug('%s deploy block not cached, fetching from chain', address)

@a_sync_db_session
def set_deploy_block(address: str, deploy_block: int) -> None:
    from y._db.utils._ep import _get_get_token
    get_token = _get_get_token()
    ensure_block(deploy_block, sync=True)
    get_token(address, sync=True).deploy_block = (chain.id, deploy_block)
    logger.debug('deploy block cached for %s: %s', address, deploy_block)




import asyncio
from contextlib import suppress
from decimal import Decimal, InvalidOperation
from typing import List, Optional

from brownie import chain

from y import constants
from y._db.entities import Price, insert
from y._db.utils.decorators import a_sync_db_session
from y._db.utils.token import ensure_token
from y._db.utils.utils import ensure_block


@a_sync_db_session
def get_price(address: str, block: int) -> Optional[Decimal]:
    ensure_block(block, sync=True)
    ensure_token(address)
    if address == constants.EEE_ADDRESS:
        address = constants.WRAPPED_GAS_COIN
    if price := Price.get(
        token = (chain.id, address), 
        block = (chain.id, block), 
    ):
        return price.price

async def set_price(address: str, block: int, price: Decimal) -> None:
    _tasks.append(asyncio.create_task(coro=_set_price(address, block, price), name=f"set_price {price} for {address} at {block}"))
    for t in _tasks[:]:
        if t.done():
            await t
            _tasks.remove(t)

@a_sync_db_session
def _set_price(address: str, block: int, price: Decimal) -> None:
    with suppress(InvalidOperation): # happens with really big numbers sometimes. nbd, we can just skip the cache in this case.
        ensure_block(block, sync=True)
        ensure_token(address)
        if address == constants.EEE_ADDRESS:
            address = constants.WRAPPED_GAS_COIN
        insert(
            type = Price,
            block = (chain.id, block),
            token = (chain.id, address),
            price = Decimal(price),
        )
        
_tasks: List[asyncio.Task] = []
